#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "./vec.h"

struct Vec {
  int *data;       /* dynamic memory area containing the integers */
  size_t length;   /* how many integers are currently stored in data */
  size_t capacity; /* how many integers can be stored in data */
};

Vec *vec_new() {
  Vec *xs = malloc(sizeof(Vec));
  if (xs == NULL) {
    return NULL;
  }
  xs->capacity = 1;
  xs->length = 0;
  xs->data = malloc(xs->capacity * sizeof(int));
  if (xs->data == NULL) {
    free(xs);
    return NULL;
  }
  return xs;
}

void vec_free(Vec *xs) {
  free(xs->data);
  free(xs);
}

int *vec_at(Vec *xs, size_t i) { return xs->data + i; }

size_t vec_length(Vec *xs) { return xs->length; }

size_t vec_capacity(Vec *xs) { return xs->capacity; }

bool vec_set_capacity(Vec *xs, size_t capacity) {
  if (xs->length <= capacity) {
    xs->capacity = capacity;
    xs->data = realloc(xs->data, capacity * sizeof(int));
    return xs->data != NULL;
  }
  return true;
}

bool vec_push(Vec *xs, int x) {
  if (xs->capacity == xs->length) {
    bool success = vec_set_capacity(xs, xs->capacity * 2);
    if (!success) {
      return false;
    }
  }
  xs->data[xs->length] = x;
  xs->length++;
  return true;
};

void vec_pop(Vec *xs) {
  if (xs->length > 0) {
    xs->length--;
    if (xs->length * 2 == xs->capacity) {
      vec_set_capacity(xs, xs->length);
    }
  }
}

void swap_int(int *x, int *y) {
  int tmp = *x;
  *x = *y;
  *y = tmp;
}

int *vec_min_between(Vec *xs, size_t from, size_t to) {
  if (from >= to || vec_length(xs) < to) {
    return NULL;
  }
  int *min = vec_at(xs, from);
  for (size_t i = from + 1; i < to; ++i) {
    int *candidate = vec_at(xs, i);
    if (*candidate < *min) {
      min = candidate;
    }
  }
  return min;
}

void vec_sort(Vec *xs) {
  size_t len = vec_length(xs);
  for (size_t i = 0; i < len; ++i) {
    swap_int(vec_min_between(xs, i, len), vec_at(xs, i));
  }
}

void vec_print(Vec *xs) {
  printf("Vector at address %p has %ld elements and capacity %ld.\n", xs,
         vec_length(xs), vec_capacity(xs));
  for (size_t i = 0; i < vec_length(xs); ++i) {
    int *xi = vec_at(xs, i);
    printf("  vec[%ld] = %d (address %p)\n", i, *xi, xi);
  }
  printf("\n");
}
